# pylint: disable=C0103,R0902,C0301
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import dataclasses
import typing

import octobot_commons.profiles.profile as profile_import
import octobot_commons.dataclasses
import octobot_commons.constants as constants


@dataclasses.dataclass
class ProfileDetailsData(octobot_commons.dataclasses.FlexibleDataclass):
    name: str = ""
    id: typing.Union[str, None] = None
    bot_id: typing.Union[str, None] = None
    version: typing.Union[str, None] = None
    user_id: typing.Union[str, None] = None
    nested_strategy_config_id: typing.Union[str, None] = None


@dataclasses.dataclass
class CryptoCurrencyData(octobot_commons.dataclasses.FlexibleDataclass):
    trading_pairs: list[str]
    name: typing.Union[str, None] = None
    enabled: bool = True


@dataclasses.dataclass
class ExchangeData(octobot_commons.dataclasses.FlexibleDataclass):
    exchange_credential_id: typing.Union[str, None] = (
        None  # deprecated, use exchange_account_id instead, exchange_credential_id is updated in exchange_data.auth_details
    )
    internal_name: typing.Union[str, None] = None
    exchange_type: str = constants.DEFAULT_EXCHANGE_TYPE
    exchange_id: typing.Union[str, None] = None
    exchange_account_id: typing.Union[str, None] = None
    sandboxed: bool = False


@dataclasses.dataclass
class FutureSymbolData(octobot_commons.dataclasses.FlexibleDataclass):
    symbol: str
    leverage: typing.Union[float, None] = None


@dataclasses.dataclass
class FutureExchangeData(octobot_commons.dataclasses.FlexibleDataclass):
    default_leverage: typing.Union[float, None] = None
    symbol_data: list[FutureSymbolData] = dataclasses.field(default_factory=list)

    # pylint: disable=E1134
    def __post_init__(self):
        if self.symbol_data and isinstance(self.symbol_data[0], dict):
            self.symbol_data = [
                FutureSymbolData.from_dict(symbol_datum)
                for symbol_datum in self.symbol_data
            ]


@dataclasses.dataclass
class TraderData(octobot_commons.dataclasses.FlexibleDataclass):
    enabled: bool = True


@dataclasses.dataclass
class TraderSimulatorData(octobot_commons.dataclasses.FlexibleDataclass):
    enabled: bool = False
    starting_portfolio: dict[str, float] = dataclasses.field(default_factory=dict)
    maker_fees: float = 0.1
    taker_fees: float = 0.1


@dataclasses.dataclass
class MinimalFund(octobot_commons.dataclasses.FlexibleDataclass):
    asset: str
    available: float
    total: float

    @classmethod
    def from_dict(cls, dict_value: dict):
        to_use_dict = copy.copy(dict_value)
        if "value" in dict_value:
            if "available" not in dict_value:
                to_use_dict["available"] = dict_value["value"]
            if "total" not in dict_value:
                to_use_dict["total"] = dict_value["value"]
        return super().from_dict(to_use_dict)


@dataclasses.dataclass
class TradingData(octobot_commons.dataclasses.FlexibleDataclass):
    reference_market: str
    minimal_funds: list[MinimalFund] = dataclasses.field(default_factory=list)
    risk: float = 1.0
    sub_portfolio: dict[str, float] = dataclasses.field(default_factory=dict)
    sellable_assets: typing.Optional[list[str]] = None

    # pylint: disable=E1134
    def __post_init__(self):
        if self.minimal_funds and isinstance(self.minimal_funds[0], dict):
            self.minimal_funds = [
                MinimalFund.from_dict(minimal_fund)
                for minimal_fund in self.minimal_funds
            ]


@dataclasses.dataclass
class TentaclesData(octobot_commons.dataclasses.FlexibleDataclass):
    name: str
    config: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class BacktestingContext(octobot_commons.dataclasses.FlexibleDataclass):
    start_time_delta: float = 0
    update_interval: float = 7 * constants.DAYS_TO_SECONDS
    starting_portfolio: dict = dataclasses.field(default_factory=dict)
    exchanges: list[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class OptionsData(octobot_commons.dataclasses.FlexibleDataclass):
    values: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ProfileData(octobot_commons.dataclasses.MinimizableDataclass):
    profile_details: ProfileDetailsData
    crypto_currencies: list[CryptoCurrencyData]
    trading: TradingData
    exchanges: list[ExchangeData] = dataclasses.field(default_factory=list)
    future_exchange_data: FutureExchangeData = dataclasses.field(
        default_factory=FutureExchangeData
    )
    trader: TraderData = dataclasses.field(default_factory=TraderData)
    trader_simulator: TraderSimulatorData = dataclasses.field(
        default_factory=TraderSimulatorData
    )
    tentacles: list[TentaclesData] = dataclasses.field(default_factory=list)
    backtesting_context: BacktestingContext = dataclasses.field(
        default_factory=BacktestingContext
    )
    options: OptionsData = dataclasses.field(default_factory=OptionsData)

    # pylint: disable=E1134
    def __post_init__(self):
        if self.crypto_currencies and isinstance(self.crypto_currencies[0], dict):
            self.crypto_currencies = [
                CryptoCurrencyData.from_dict(crypto_currency)
                for crypto_currency in self.crypto_currencies
            ]
        if self.exchanges and isinstance(self.exchanges[0], dict):
            self.exchanges = [
                ExchangeData.from_dict(exchange) for exchange in self.exchanges
            ]
        if self.tentacles and isinstance(self.tentacles[0], dict):
            self.tentacles = (
                [TentaclesData.from_dict(tentacle) for tentacle in self.tentacles]
                if self.tentacles
                else []
            )

    @classmethod
    def from_profile(cls, profile: profile_import.Profile):
        """
        Creates a cls instance from the given profile
        """
        profile_dict = profile.as_dict()
        content = profile_dict[constants.PROFILE_CONFIG]
        return cls.from_dict(
            {
                "profile_details": {
                    "id": profile_dict[constants.CONFIG_PROFILE][constants.CONFIG_ID],
                    "name": profile_dict[constants.CONFIG_PROFILE][
                        constants.CONFIG_NAME
                    ],
                },
                "crypto_currencies": [
                    {
                        "trading_pairs": details.get(constants.CONFIG_CRYPTO_PAIRS, []),
                        "name": currency,
                        "enabled": details.get(constants.CONFIG_ENABLED_OPTION, True),
                    }
                    for currency, details in content[
                        constants.CONFIG_CRYPTO_CURRENCIES
                    ].items()
                ],
                "trader": {
                    "enabled": content[constants.CONFIG_TRADER][
                        constants.CONFIG_ENABLED_OPTION
                    ],
                },
                "trader_simulator": {
                    "enabled": content[constants.CONFIG_SIMULATOR][
                        constants.CONFIG_ENABLED_OPTION
                    ],
                    "starting_portfolio": content[constants.CONFIG_SIMULATOR][
                        constants.CONFIG_STARTING_PORTFOLIO
                    ],
                    "maker_fees": content[constants.CONFIG_SIMULATOR][
                        constants.CONFIG_SIMULATOR_FEES
                    ].get(constants.CONFIG_SIMULATOR_FEES_MAKER, 0.0),
                    "taker_fees": content[constants.CONFIG_SIMULATOR][
                        constants.CONFIG_SIMULATOR_FEES
                    ].get(constants.CONFIG_SIMULATOR_FEES_TAKER, 0.0),
                },
                "trading": {
                    "reference_market": content[constants.CONFIG_TRADING][
                        constants.CONFIG_TRADER_REFERENCE_MARKET
                    ],
                    "risk": content[constants.CONFIG_TRADING][
                        constants.CONFIG_TRADER_RISK
                    ],
                },
                "tentacles": [],
            }
        )

    def to_profile(self, to_create_profile_path: str) -> profile_import.Profile:
        """
        Returns a new Profile from self
        """
        profile = profile_import.Profile(to_create_profile_path)
        profile.from_dict(self._to_profile_dict())
        return profile

    def set_tentacles_config(self, config_by_tentacle: dict):
        """
        Update self.tentacles from the given config_by_tentacle
        """
        self.tentacles = [
            TentaclesData(name=tentacle, config=config)
            for tentacle, config in config_by_tentacle.items()
        ]

    def _to_profile_dict(self) -> dict:
        return {
            constants.PROFILE_CONFIG: {
                constants.CONFIG_CRYPTO_CURRENCIES: {
                    crypto_currency.name: {
                        constants.CONFIG_CRYPTO_PAIRS: crypto_currency.trading_pairs,
                        constants.CONFIG_ENABLED_OPTION: crypto_currency.enabled,
                    }
                    for crypto_currency in self.crypto_currencies
                },
                constants.CONFIG_EXCHANGES: {
                    exchange_details.internal_name: {
                        constants.CONFIG_ENABLED_OPTION: True,
                        constants.CONFIG_EXCHANGE_TYPE: exchange_details.exchange_type,
                    }
                    for exchange_details in self.exchanges
                },
                constants.CONFIG_TRADER: {
                    constants.CONFIG_ENABLED_OPTION: self.trader.enabled,
                    constants.CONFIG_LOAD_TRADE_HISTORY: True,
                },
                constants.CONFIG_SIMULATOR: {
                    constants.CONFIG_ENABLED_OPTION: self.trader_simulator.enabled,
                    constants.CONFIG_STARTING_PORTFOLIO: self.trader_simulator.starting_portfolio
                    or (
                        self.backtesting_context.starting_portfolio
                        if self.backtesting_context
                        else {}
                    ),
                    constants.CONFIG_SIMULATOR_FEES: {
                        constants.CONFIG_SIMULATOR_FEES_MAKER: self.trader_simulator.maker_fees,
                        constants.CONFIG_SIMULATOR_FEES_TAKER: self.trader_simulator.taker_fees,
                    },
                },
                constants.CONFIG_TRADING: {
                    constants.CONFIG_TRADER_REFERENCE_MARKET: self.trading.reference_market,
                    constants.CONFIG_TRADER_RISK: self.trading.risk,
                },
            },
            constants.CONFIG_PROFILE: dataclasses.asdict(self.profile_details),
        }



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.symbols


def test_parse_symbol():
    assert octobot_commons.symbols.parse_symbol("BTC/USDT") == octobot_commons.symbols.Symbol("BTC/USDT")


def test_merge_symbol():
    assert octobot_commons.symbols.merge_symbol("BTC/USDT") == "BTCUSDT"
    assert octobot_commons.symbols.merge_symbol("BTC/USDT:USDT") == "BTCUSDT_USDT"


def test_merge_currencies():
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT") == "BTC/USDT"
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT", "BTC") == "BTC/USDT:BTC"
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT", settlement_asset="XXX", market_separator="g", settlement_separator="d") == "BTCgUSDTdXXX"
    assert (
        octobot_commons.symbols.merge_currencies(
            "will-bitcoin-replace-sha-256-before-2027",
            "USDC",
            "USDC",
            "261231",
            "0",
            octobot_commons.enums.OptionTypes.PUT.value
        )
        == "will-bitcoin-replace-sha-256-before-2027/USDC:USDC-261231-0-P"
    )
    assert (
        octobot_commons.symbols.merge_currencies(
            "will-bitcoin-replace-sha-256-before-2027",
            "USDC",
            "USDC",
            "261231",
            "0",
            "test"
        )
        == "will-bitcoin-replace-sha-256-before-2027/USDC:USDC-261231-0-TEST"
    )
    assert (
        octobot_commons.symbols.merge_currencies(
            "will-bitcoin-replace-sha-256-before-2027",
            "USDC",
            "USDC",
            "261231",
            "0",
            None
        )
        == "will-bitcoin-replace-sha-256-before-2027/USDC:USDC"
    )


def test_convert_symbol():
    assert octobot_commons.symbols.convert_symbol("BTC-USDT", symbol_separator="-") == "BTC/USDT"


def test_is_symbol():
    # Test with default separator (/)
    assert octobot_commons.symbols.is_symbol("BTC/USDT") is True
    assert octobot_commons.symbols.is_symbol("ETH/USDT") is True
    assert octobot_commons.symbols.is_symbol("BTC/USDT:USDT") is True
    assert octobot_commons.symbols.is_symbol("BTC") is False
    assert octobot_commons.symbols.is_symbol("USDT") is False
    assert octobot_commons.symbols.is_symbol("ETH") is False
    
    # Test with custom separator (-)
    assert octobot_commons.symbols.is_symbol("BTC-USDT", separator="-") is True
    assert octobot_commons.symbols.is_symbol("ETH-USDT", separator="-") is True
    assert octobot_commons.symbols.is_symbol("BTC", separator="-") is False
    assert octobot_commons.symbols.is_symbol("USDT", separator="-") is False
    
    # Test with custom separator (:)
    assert octobot_commons.symbols.is_symbol("BTC/USDT:USDT", separator=":") is True
    assert octobot_commons.symbols.is_symbol("BTC/USDT", separator=":") is False
    assert octobot_commons.symbols.is_symbol("BTC", separator=":") is False
    
    # Test with custom separator (|)
    assert octobot_commons.symbols.is_symbol("BTC|USDT", separator="|") is True
    assert octobot_commons.symbols.is_symbol("BTC", separator="|") is False
    
    # Test edge cases
    assert octobot_commons.symbols.is_symbol("", separator="/") is False
    assert octobot_commons.symbols.is_symbol("/", separator="/") is True
    assert octobot_commons.symbols.is_symbol("BTC/USDT/ETH", separator="/") is True



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.number_util import round_into_str_with_max_digits, round_into_float_with_max_digits


def test_round_into_max_digits():
    assert round_into_str_with_max_digits(125.0256, 2) == '125.03'
    assert round_into_float_with_max_digits(125.0210, 2) == 125.02
    assert round_into_float_with_max_digits(1301, 5) == 1301.00000
    assert round_into_float_with_max_digits(59866, 0) == 59866
    assert round_into_float_with_max_digits(1.567824117582484154178, 15) == 1.567824117582484
    assert round_into_float_with_max_digits(0.000000059, 8) == 0.00000006
    assert not round_into_float_with_max_digits(8712661000.1273185137283, 10) == 8712661000.127318
    assert round_into_float_with_max_digits(8712661000.1273185137283, 10) == 8712661000.1273185
    assert round_into_float_with_max_digits(8712661000.1273185137283, 10) == 8712661000.12731851
    assert round_into_float_with_max_digits(8712661000.1273185137283, 10) == 8712661000.127318513
    assert round_into_float_with_max_digits(8712661000.1273185137283, 10) == 8712661000.1273185137
    assert round_into_float_with_max_digits(0.0000000000001, 5) == 0
    assert not round_into_float_with_max_digits(0.0000000000001, 13) == 0




# pylint: disable=R0801,R0401
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.dsl_interpreter.interpreter import Interpreter
from octobot_commons.dsl_interpreter.operator import (
    Operator,
    OperatorParameterType,
    ComputedOperatorParameterType,
)
from octobot_commons.dsl_interpreter.dictionnaries import (
    get_all_operators,
    clear_get_all_operators_cache,
)
from octobot_commons.dsl_interpreter.operator_parameter import OperatorParameter
from octobot_commons.dsl_interpreter.operator_docs import OperatorDocs
from octobot_commons.dsl_interpreter.operators import (
    BinaryOperator,
    UnaryOperator,
    CompareOperator,
    NaryOperator,
    CallOperator,
    NameOperator,
    ExpressionOperator,
)
from octobot_commons.dsl_interpreter.interpreter_dependency import InterpreterDependency

__all__ = [
    "get_all_operators",
    "clear_get_all_operators_cache",
    "Interpreter",
    "Operator",
    "OperatorParameter",
    "OperatorDocs",
    "BinaryOperator",
    "UnaryOperator",
    "CompareOperator",
    "NaryOperator",
    "CallOperator",
    "NameOperator",
    "ExpressionOperator",
    "InterpreterDependency",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import numpy as np

import octobot_commons.errors
import octobot_commons.constants
import octobot_commons.dsl_interpreter.interpreter_dependency as dsl_interpreter_dependency
import octobot_commons.dsl_interpreter.operator_parameter as dsl_interpreter_operator_parameter
import octobot_commons.dsl_interpreter.operator_docs as dsl_interpreter_operator_docs

OperatorParameterType = typing.Union[
    str, int, float, bool, None, list, np.ndarray, "Operator"
]
ComputedOperatorParameterType = typing.Union[
    str, int, float, bool, None, list, np.ndarray
]


class Operator:
    """
    Operator class is used to represent an operator in the DSL.
    It can have one or more parameters which are used to compute the result of the operator.
    """

    MIN_PARAMS: typing.Optional[int] = (
        None  # min number of parameters when not defined in get_parameters()
    )
    MAX_PARAMS: typing.Optional[int] = (
        None  # max number of parameters when not defined in get_parameters()
    )
    NAME: str = (
        ""  # name of the operator as written in the DSL expression, if not provided, get_name() will be used
    )
    DESCRIPTION: str = ""  # description of the operator
    EXAMPLE: str = ""  # example of the operator in the DSL

    def __init__(self, *parameters: OperatorParameterType, **kwargs: typing.Any):
        self._validate_parameters(parameters)
        self.parameters = parameters
        self.kwargs = kwargs

    @staticmethod
    def get_name() -> str:
        """
        Get the name of the operator, as seen in the AST parsed expression.
        """
        raise NotImplementedError("get_name is not implemented")

    @staticmethod
    def get_library() -> str:
        """
        Get the library of the operator.
        """
        return octobot_commons.constants.BASE_OPERATORS_LIBRARY

    def _validate_parameters(
        self, parameters: typing.List[OperatorParameterType]
    ) -> None:
        """
        Validate the parameters of the operator.
        """
        if self.MIN_PARAMS is not None and len(parameters) < self.MIN_PARAMS:
            raise octobot_commons.errors.InvalidParametersError(
                f"{self.get_name()} requires at least {self.MIN_PARAMS} parameter(s)"
            )
        if self.MAX_PARAMS is not None and len(parameters) > self.MAX_PARAMS:
            raise octobot_commons.errors.InvalidParametersError(
                f"{self.get_name()} supports up to {self.MAX_PARAMS} parameters"
            )
        if expected_parameters := self.get_parameters():
            min_params = len(tuple(p for p in expected_parameters if p.required))
            max_params = len(tuple(p for p in expected_parameters))
            if len(parameters) < min_params:
                raise octobot_commons.errors.InvalidParametersError(
                    f"{self.get_name()} requires at least {min_params} "
                    f"parameter(s): {self.get_parameters_description()}"
                )
            if max_params is not None and len(parameters) > max_params:
                raise octobot_commons.errors.InvalidParametersError(
                    f"{self.get_name()} supports up to {max_params} "
                    f"parameters: {self.get_parameters_description()}"
                )

    @classmethod
    def get_parameters_description(cls) -> str:
        """
        Get the description of the parameters of the operator.
        """
        return ", ".join(
            (f"{i+1}: {param}" for i, param in enumerate(cls.get_parameters()))
        )

    @classmethod
    def get_docs(cls) -> dsl_interpreter_operator_docs.OperatorDocs:
        """
        Get the documentation of the operator.
        """
        return dsl_interpreter_operator_docs.OperatorDocs(
            name=cls.NAME or cls.get_name(),
            description=cls.DESCRIPTION,
            type=cls.get_library(),
            example=cls.EXAMPLE,
            parameters=cls.get_parameters(),
        )

    @staticmethod
    def get_parameters() -> list[dsl_interpreter_operator_parameter.OperatorParameter]:
        """
        return: the description of the parameters of the operator.
        """
        return []

    async def pre_compute(self) -> None:  # rename pre_compute
        """
        Refreshes the operator data, override if necessary.
        Will always be called before compute()
        """
        for parameter in self.parameters:
            if isinstance(parameter, Operator):
                await parameter.pre_compute()

    def compute(self) -> ComputedOperatorParameterType:
        """
        Compute the result of the operator considering its computed parameters.
        """
        raise NotImplementedError("compute is not implemented")

    def get_computed_parameters(self) -> typing.List[ComputedOperatorParameterType]:
        """
        Get the computed parameters of the operator.
        Here computed means that any nested operator has already been computed.
        """
        return [
            parameter.compute() if isinstance(parameter, Operator) else parameter
            for parameter in self.parameters
        ]

    def get_dependencies(
        self,
    ) -> typing.List[dsl_interpreter_dependency.InterpreterDependency]:
        """
        Get the dependencies of the operator.
        """
        dependencies = []
        for parameter in self.parameters:
            if isinstance(parameter, Operator):
                dependencies.extend(parameter.get_dependencies())
        return dependencies



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.symbols import symbol_util

from octobot_commons.symbols.symbol_util import (
    parse_symbol,
    merge_symbol,
    merge_currencies,
    convert_symbol,
    is_symbol,
    is_usd_like_coin,
    get_most_common_usd_like_symbol,
)

from octobot_commons.symbols import symbol

from octobot_commons.symbols.symbol import (
    Symbol,
)


__all__ = [
    "parse_symbol",
    "merge_symbol",
    "merge_currencies",
    "convert_symbol",
    "is_symbol",
    "is_usd_like_coin",
    "get_most_common_usd_like_symbol",
    "Symbol",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import traceback

import octobot_commons.html_util as html_util


def test_summarize_page_content_invalid_content():
    assert html_util.summarize_page_content("plop") == [('message', 'plop')]
    with pytest.raises(TypeError):
        html_util.summarize_page_content(1234)
    with pytest.raises(TypeError):
        html_util.summarize_page_content(RuntimeError("124"))
    with pytest.raises(TypeError):
        html_util.summarize_page_content({})
    assert html_util.summarize_page_content(RATE_LIMIT_INVALID_HTMLS[0]) == []
    with pytest.raises(IndexError):
        html_util.summarize_page_content(RATE_LIMIT_INVALID_HTMLS[1])
    with pytest.raises(IndexError):
        html_util.summarize_page_content(RATE_LIMIT_INVALID_HTMLS[2])


def test_summarize_page_content_valid_content():
    assert html_util.summarize_page_content(RATE_LIMIT_HTML) == [
        ('title', 'Access denied | api.hollaex.com used Cloudflare to restrict access'),
        ('div', 'Please enable cookies.'),
        ('span', 'Error'),
        ('span', '1006'),
        ('span', 'Ray ID: 8e61e5180c749e8a •'),
        ('span', '2024-11-21 16:00:49 UTC'),
        ('h2', 'Access denied'),
        ('h2', 'What happened?'),
        ('p', 'The owner of this website (api.hollaex.com) has banned your IP address (123.123.123.123).'),
        ('div', 'Was this page helpful?'),
        ('div', 'Thank you for your feedback!'),
        ('span', 'Cloudflare Ray ID:'),
        ('strong', '8e61e5180c749e8a'),
        ('span', 'Your IP:'),
        ('span', '123.123.123.123'),
        ('span', 'Performance & security by'),
        ('a', 'Cloudflare')
    ]
    assert html_util.summarize_page_content(RATE_LIMIT_HTML_WITH_MESSAGE) == [
        ('message', 'hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 Forbidden'),
        ('title', 'Access denied | api.hollaex.com used Cloudflare to restrict access'),
        ('div', 'Please enable cookies.'),
        ('span', 'Error'),
        ('span', '1006'),
        ('span', 'Ray ID: 8e61e5180c749e8a •'),
        ('span', '2024-11-21 16:00:49 UTC'),
        ('h2', 'Access denied'),
        ('h2', 'What happened?'),
        ('p', 'The owner of this website (api.hollaex.com) has banned your IP address (123.123.123.123).'),
        ('div', 'Was this page helpful?'),
        ('div', 'Thank you for your feedback!'),
        ('span', 'Cloudflare Ray ID:'),
        ('strong', '8e61e5180c749e8a'),
        ('span', 'Your IP:'),
        ('span', '123.123.123.123'),
        ('span', 'Performance & security by'),
        ('a', 'Cloudflare')
    ]
    assert html_util.summarize_page_content(RATE_LIMIT_HTML_WITH_MESSAGE, max_element_text_size=4) == [
        ('message', 'holl[...]'),
        ('title', 'Acce[...]'),
        ('div', 'Plea[...]'),
        ('span', 'Erro[...]'),
        ('span', '1006'),
        ('span', 'Ray [...]'),
        ('span', '2024[...]'),
        ('h2', 'Acce[...]'),
        ('h2', 'What[...]'),
        ('p', 'The [...]'),
        ('div', 'Was [...]'),
        ('div', 'Than[...]'),
        ('span', 'Clou[...]'),
        ('strong', '8e61[...]'),
        ('span', 'Your[...]'),
        ('span', '123.[...]'),
        ('span', 'Perf[...]'),
        ('a', 'Clou[...]')
    ]


def test_pretty_print_summary():
    assert html_util.pretty_print_summary(html_util.summarize_page_content(RATE_LIMIT_HTML_WITH_MESSAGE)) == (
        'message<hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 '
        'Forbidden>; title<Access denied | api.hollaex.com used Cloudflare to '
        'restrict access>; div<Please enable cookies.>; span<Error>; span<1006>; '
        'span<Ray ID: 8e61e5180c749e8a •>; span<2024-11-21 16:00:49 UTC>; h2<Access '
        'denied>; h2<What happened?>; p<The owner of this website (api.hollaex.com) '
        'has banned your IP address (123.123.123.123).>; div<Was this page helpful?>; '
        'div<Thank you for your feedback!>; span<Cloudflare Ray ID:>; '
        'strong<8e61e5180c749e8a>; span<Your IP:>; span<123.123.123.123>; '
        'span<Performance & security by>; a<Cloudflare>'
    )


def test_is_html_content():
    assert html_util.is_html_content(RATE_LIMIT_HTML_WITH_MESSAGE) is True
    assert html_util.is_html_content(RATE_LIMIT_HTML) is True
    assert all(html_util.is_html_content(html) for html in RATE_LIMIT_INVALID_HTMLS)
    assert html_util.is_html_content("plpoop") is False
    assert html_util.is_html_content("<html>") is False
    assert html_util.is_html_content("</html/>") is False
    assert html_util.is_html_content("<html/>") is False
    assert html_util.is_html_content("</html>") is True


def test_get_html_summary_if_relevant():
    # valid html
    assert html_util.get_html_summary_if_relevant(RATE_LIMIT_HTML_WITH_MESSAGE) == (
        'message<hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 '
        'Forbidden>; title<Access denied | api.hollaex.com used Cloudflare to '
        'restrict access>; div<Please enable cookies.>; span<Error>; span<1006>; '
        'span<Ray ID: 8e61e5180c749e8a •>; span<2024-11-21 16:00:49 UTC>; h2<Access '
        'denied>; h2<What happened?>; p<The owner of this website (api.hollaex.com) '
        'has banned your IP address (123.123.123.123).>; div<Was this page helpful?>; '
        'div<Thank you for your feedback!>; span<Cloudflare Ray ID:>; '
        'strong<8e61e5180c749e8a>; span<Your IP:>; span<123.123.123.123>; '
        'span<Performance & security by>; a<Cloudflare>'
    )
    # invalid html
    assert html_util.get_html_summary_if_relevant("PLOPOPDS") == "PLOPOPDS"
    # non string
    assert html_util.get_html_summary_if_relevant(123) == "123"
    assert html_util.get_html_summary_if_relevant(Exception("plop")) == "plop"
    # raising error does not propagate
    with mock.patch.object(html_util, "summarize_page_content", mock.Mock(side_effect=Exception("err"))) \
        as summarize_page_content_mock:
        # not html
        assert html_util.get_html_summary_if_relevant("PLOPOPDS") == "PLOPOPDS"
        summarize_page_content_mock.assert_not_called()
        # html
        assert html_util.get_html_summary_if_relevant(
            "dfdsfsdfsd</html>", max_element_text_size=3
        ) == "dfdsfsdfsd</html>"
        summarize_page_content_mock.assert_called_once_with(
            "dfdsfsdfsd</html>", max_element_text_size=3
        )
    # raising error does not propagate
    with mock.patch.object(html_util, "pretty_print_summary", mock.Mock(side_effect=Exception("err"))) \
        as pretty_print_summary_mock:
        # not html
        assert html_util.get_html_summary_if_relevant("PLOPOPDS") == "PLOPOPDS"
        pretty_print_summary_mock.assert_not_called()
        # html
        assert html_util.get_html_summary_if_relevant("<html>dfdsfsdfsd</html>") == "<html>dfdsfsdfsd</html>"
        pretty_print_summary_mock.assert_called_once_with(
            [('html', 'dfdsfsdfsd')]
        )


def test_summarize_exception_html_cause_if_relevant():
    assert html_util.summarize_exception_html_cause_if_relevant(Exception("plop")) is None
    try:
        # lvl. 1
        try:
            # lvl. 2
            try:
                # lvl. 3
                try:
                    # lvl. 4
                    raise IndexError(RATE_LIMIT_HTML_WITH_MESSAGE)
                except IndexError as err_4:
                    assert err_4.__cause__ is None
                    assert "<script>" in str(err_4)
                    str_traceback = traceback.format_exc()
                    assert "<script>" in str_traceback
                    # does not crash if __cause__ is None
                    html_util.summarize_exception_html_cause_if_relevant(err_4)
                    str_traceback = traceback.format_exc()
                    assert "<script>" not in str_traceback
                    raise KeyError(html_util.get_html_summary_if_relevant(err_4)) from err_4
            except KeyError as err_3:
                # lvl. 3
                assert err_3.args == (
                    'message<hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 '
                    'Forbidden>; title<Access denied | api.hollaex.com used Cloudflare to '
                    'restrict access>; div<Please enable cookies.>; span<Error>; span<1006>; '
                    'span<Ray ID: 8e61e5180c749e8a •>; span<2024-11-21 16:00:49 UTC>; h2<Access '
                    'denied>; h2<What happened?>; p<The owner of this website (api.hollaex.com) '
                    'has banned your IP address (123.123.123.123).>; div<Was this page helpful?>; '
                    'div<Thank you for your feedback!>; span<Cloudflare Ray ID:>; '
                    'strong<8e61e5180c749e8a>; span<Your IP:>; span<123.123.123.123>; '
                    'span<Performance & security by>; a<Cloudflare>',
                )
                # cause is not summarized
                assert "<script>" not in str(err_3.__cause__.args)
                html_util.summarize_exception_html_cause_if_relevant(err_3)
                # cause has been summarized
                assert err_3.__cause__.args == (
                    'message<hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 '
                    'Forbidden>; title<Access denied | api.hollaex.com used Cloudflare to '
                    'restrict access>; div<Please enable cookies.>; span<Error>; span<1006>; '
                    'span<Ray ID: 8e61e5180c749e8a •>; span<2024-11-21 16:00:49 UTC>; h2<Access '
                    'denied>; h2<What happened?>; p<The owner of this website (api.hollaex.com) '
                    'has banned your IP address (123.123.123.123).>; div<Was this page helpful?>; '
                    'div<Thank you for your feedback!>; span<Cloudflare Ray ID:>; '
                    'strong<8e61e5180c749e8a>; span<Your IP:>; span<123.123.123.123>; '
                    'span<Performance & security by>; a<Cloudflare>',
                )
                raise NotImplementedError(err_3) from err_3
        except NotImplementedError as err_2:
            # lvl. 2
            assert err_2.args[0].args == (
                'message<hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 '
                'Forbidden>; title<Access denied | api.hollaex.com used Cloudflare to '
                'restrict access>; div<Please enable cookies.>; span<Error>; span<1006>; '
                'span<Ray ID: 8e61e5180c749e8a •>; span<2024-11-21 16:00:49 UTC>; h2<Access '
                'denied>; h2<What happened?>; p<The owner of this website (api.hollaex.com) '
                'has banned your IP address (123.123.123.123).>; div<Was this page helpful?>; '
                'div<Thank you for your feedback!>; span<Cloudflare Ray ID:>; '
                'strong<8e61e5180c749e8a>; span<Your IP:>; span<123.123.123.123>; '
                'span<Performance & security by>; a<Cloudflare>',
            )
            assert err_2.__cause__.__cause__.__class__ == IndexError
            assert err_2.__cause__.__class__ == KeyError
            raise ZeroDivisionError from err_2
    except ZeroDivisionError as err_1:
        # lvl. 1
        assert "<script>" not in str(err_1)
        str_traceback = traceback.format_exc()
        assert "<script>" not in str_traceback

    # no summary at lower levels
    try:
        # lvl. 1
        try:
            # lvl. 2
            try:
                # lvl. 3
                try:
                    # lvl. 4
                    raise IndexError(RATE_LIMIT_HTML_WITH_MESSAGE)
                except IndexError as err_4:
                    raise KeyError(html_util.get_html_summary_if_relevant(err_4)) from err_4
            except KeyError as err_3:
                # lvl. 3
                raise NotImplementedError(err_3) from err_3
        except NotImplementedError as err_2:
            # lvl. 2
            raise ZeroDivisionError from err_2
    except ZeroDivisionError as err_1:
        # lvl. 1
        # summarized in error message
        assert "<script>" not in str(err_1)
        # not yet summarized in causes (used by traceback)
        str_traceback = traceback.format_exc()
        assert "<script>" in str_traceback

        # summarize
        html_util.summarize_exception_html_cause_if_relevant(err_1)
        str_traceback = traceback.format_exc()
        assert "<script>" not in str_traceback


RATE_LIMIT_INVALID_HTMLS = [
"""
<!DOCTYPE html>
<!--[if lt IE 7]> <html class="no-js ie6 oldie" lang="en-US"> <![endif]-->
<!--[if IE 7]>    <html class="no-js ie7 oldie" lang="en-US"> <![endif]-->
<!--[if IE 8]>    <html class="no-js ie8 oldie" lang="en-US"> <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="en-US"> <!--<![endif]-->
<head>
</html>
""",
"""
PLOP
</html>
""",
"""
</ll>
PLOP
</html>
""",
]

RATE_LIMIT_HTML = """
<!DOCTYPE html>
<!--[if lt IE 7]> <html class="no-js ie6 oldie" lang="en-US"> <![endif]-->
<!--[if IE 7]>    <html class="no-js ie7 oldie" lang="en-US"> <![endif]-->
<!--[if IE 8]>    <html class="no-js ie8 oldie" lang="en-US"> <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="en-US"> <!--<![endif]-->
<head>
<title>Access denied | api.hollaex.com used Cloudflare to restrict access</title>
<meta charset="UTF-8" />
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
<meta http-equiv="X-UA-Compatible" content="IE=Edge" />
<meta name="robots" content="noindex, nofollow" />
<meta name="viewport" content="width=device-width,initial-scale=1" />
<link rel="stylesheet" id="cf_styles-css" href="/cdn-cgi/styles/main.css" />


<script>
(function(){if(document.addEventListener&&window.XMLHttpRequest&&JSON&&JSON.stringify){var e=function(a){var c=document.getElementById("error-feedback-survey"),d=document.getElementById("error-feedback-success"),b=new XMLHttpRequest;a={event:"feedback clicked",properties:{errorCode:1006,helpful:a,version:1}};b.open("POST","https://sparrow.cloudflare.com/api/v1/event");b.setRequestHeader("Content-Type","application/json");b.setRequestHeader("Sparrow-Source-Key","c771f0e4b54944bebf4261d44bd79a1e");
b.send(JSON.stringify(a));c.classList.add("feedback-hidden");d.classList.remove("feedback-hidden")};document.addEventListener("DOMContentLoaded",function(){var a=document.getElementById("error-feedback"),c=document.getElementById("feedback-button-yes"),d=document.getElementById("feedback-button-no");"classList"in a&&(a.classList.remove("feedback-hidden"),c.addEventListener("click",function(){e(!0)}),d.addEventListener("click",function(){e(!1)}))})}})();
</script>

<script defer src="https://performance.radar.cloudflare.com/beacon.js"></script>
</head>
<body>
  <div id="cf-wrapper">
    <div class="cf-alert cf-alert-error cf-cookie-error hidden" id="cookie-alert" data-translate="enable_cookies">Please enable cookies.</div>
    <div id="cf-error-details" class="p-0">
      <header class="mx-auto pt-10 lg:pt-6 lg:px-8 w-240 lg:w-full mb-15 antialiased">
         <h1 class="inline-block md:block mr-2 md:mb-2 font-light text-60 md:text-3xl text-black-dark leading-tight">
           <span data-translate="error">Error</span>
           <span>1006</span>
         </h1>
         <span class="inline-block md:block heading-ray-id font-mono text-15 lg:text-sm lg:leading-relaxed">Ray ID: 8e61e5180c749e8a &bull;</span>
         <span class="inline-block md:block heading-ray-id font-mono text-15 lg:text-sm lg:leading-relaxed">2024-11-21 16:00:49 UTC</span>
        <h2 class="text-gray-600 leading-1.3 text-3xl lg:text-2xl font-light">Access denied</h2>
      </header>

      <section class="w-240 lg:w-full mx-auto mb-8 lg:px-8">
          <div id="what-happened-section" class="w-1/2 md:w-full">
            <h2 class="text-3xl leading-tight font-normal mb-4 text-black-dark antialiased" data-translate="what_happened">What happened?</h2>
            <p>The owner of this website (api.hollaex.com) has banned your IP address (123.123.123.123).</p>

          </div>


      </section>

      <div class="feedback-hidden py-8 text-center" id="error-feedback">
    <div id="error-feedback-survey" class="footer-line-wrapper">
        Was this page helpful?
        <button class="border border-solid bg-white cf-button cursor-pointer ml-4 px-4 py-2 rounded" id="feedback-button-yes" type="button">Yes</button>
        <button class="border border-solid bg-white cf-button cursor-pointer ml-4 px-4 py-2 rounded" id="feedback-button-no" type="button">No</button>
    </div>
    <div class="feedback-success feedback-hidden" id="error-feedback-success">
        Thank you for your feedback!
    </div>
</div>


      <div class="cf-error-footer cf-wrapper w-240 lg:w-full py-10 sm:py-4 sm:px-8 mx-auto text-center sm:text-left border-solid border-0 border-t border-gray-300">
  <p class="text-13">
    <span class="cf-footer-item sm:block sm:mb-1">Cloudflare Ray ID: <strong class="font-semibold">8e61e5180c749e8a</strong></span>
    <span class="cf-footer-separator sm:hidden">&bull;</span>
    <span id="cf-footer-item-ip" class="cf-footer-item hidden sm:block sm:mb-1">
      Your IP:
      <button type="button" id="cf-footer-ip-reveal" class="cf-footer-ip-reveal-btn">Click to reveal</button>
      <span class="hidden" id="cf-footer-ip">123.123.123.123</span>
      <span class="cf-footer-separator sm:hidden">&bull;</span>
    </span>
    <span class="cf-footer-item sm:block sm:mb-1"><span>Performance &amp; security by</span> <a rel="noopener noreferrer" href="https://www.cloudflare.com/5xx-error-landing" id="brand_link" target="_blank">Cloudflare</a></span>

  </p>
  <script>(function(){function d(){var b=a.getElementById("cf-footer-item-ip"),c=a.getElementById("cf-footer-ip-reveal");b&&"classList"in b&&(b.classList.remove("hidden"),c.addEventListener("click",function(){c.classList.add("hidden");a.getElementById("cf-footer-ip").classList.remove("hidden")}))}var a=document;document.addEventListener&&a.addEventListener("DOMContentLoaded",d)})();</script>
</div><!-- /.error-footer -->


    </div><!-- /#cf-error-details -->
  </div><!-- /#cf-wrapper -->

  <script>
  window._cf_translation = {};


</script>

</body>
</html>
"""
RATE_LIMIT_HTML_WITH_MESSAGE = \
    f"hollaex GET https://api.sandbox.hollaex.com/v2/constants 403 Forbidden {RATE_LIMIT_HTML}"


#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.tentacles_management.class_inspector as class_inspector
import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.profiles.exchange_auth_data as exchange_auth_data_import
import octobot_commons.profiles.tentacles_profile_data_adapter as tentacles_profile_data_adapter


class TentaclesProfileDataTranslator:
    """
    Translates a tentacle-specific configuration into a ProfileData
    """

    def __init__(
        self,
        profile_data: profile_data_import.ProfileData,
        auth_data: list[exchange_auth_data_import.ExchangeAuthData],
    ):
        self.profile_data: profile_data_import.ProfileData = profile_data
        self.auth_data: list[exchange_auth_data_import.ExchangeAuthData] = auth_data

    async def translate(
        self,
        tentacles_data: list[profile_data_import.TentaclesData],
        additional_data: dict,
        authenticator,
        auth_key: typing.Optional[str],
    ) -> None:
        """
        updates self.profile_data by applying the given tentacles_data and
        additional_data configuration
        :param tentacles_data: the tentacles data to use
        :param additional_data: other data that can be useful in translation
        :param authenticator: authenticator to fetch data from if necessary
        :param auth_key: auth key to used if necessary
        :return:
        """
        adapter = self._get_adapter(tentacles_data)
        await adapter(tentacles_data, additional_data, authenticator, auth_key).adapt(
            self.profile_data, self.auth_data
        )

    @classmethod
    def _get_adapter(cls, tentacles_data: list[profile_data_import.TentaclesData]):
        """
        :return: the first adapter matching a TentaclesData name
        """
        adapters = cls._get_adapters()
        for tentacles_data_element in tentacles_data:
            if adapter := adapters.get(tentacles_data_element.name):
                return adapter
        raise KeyError("TentaclesData adapter not found")

    @classmethod
    def _get_adapters(cls) -> dict:
        return {
            adapter.get_tentacle_name(): adapter
            for adapter in class_inspector.get_all_classes_from_parent(
                tentacles_profile_data_adapter.TentaclesProfileDataAdapter
            )
        }



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses


import octobot_commons.dsl_interpreter.operator_parameter as dsl_interpreter_operator_parameter


@dataclasses.dataclass
class OperatorDocs:
    """
    Operator documentation class, used to store operators metadata to
    generate an operator documentation.
    """

    name: str
    description: str
    type: str
    example: str
    parameters: list[dsl_interpreter_operator_parameter.OperatorParameter]

    def to_json(self) -> dict:
        """
        Convert the operator documentation to a JSON serializable dict.
        """
        return {
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "example": self.example,
            "parameters": [parameter.to_json() for parameter in self.parameters],
        }



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import mock
import pytest
import aiohttp
import certifi

import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.constants as commons_constants

pytestmark = pytest.mark.asyncio


async def test_get_ssl_fallback_aiohttp_client_session():
    origin_where = certifi.where
    ok_get_mock_calls = []
    ko_get_mock = []

    @contextlib.asynccontextmanager
    async def _ok_get_mock(*args, **kwargs):
        ok_get_mock_calls.append(1)
        yield mock.Mock(status=200)

    @contextlib.asynccontextmanager
    async def _ko_get_mock(*args, **kwargs):
        ko_get_mock.append(1)
        yield mock.Mock(status=200)
        raise aiohttp.ClientConnectorCertificateError("ssl blabla", RuntimeError())

    with mock.patch.object(certifi, "where", mock.Mock(side_effect=origin_where)) as where_mock:

        # no need for certifi
        with mock.patch.object(aiohttp.ClientSession, "get", _ok_get_mock):
            session = await aiohttp_util.get_ssl_fallback_aiohttp_client_session(
                commons_constants.KNOWN_POTENTIALLY_SSL_FAILED_REQUIRED_URL
            )
            assert isinstance(session, aiohttp.ClientSession)
            assert len(ok_get_mock_calls) == 1
            ok_get_mock_calls.clear()
            assert len(ko_get_mock) == 0
            where_mock.assert_not_called()
            await session.close()

        # need for certifi
        with mock.patch.object(aiohttp.ClientSession, "get", _ko_get_mock):
            async with aiohttp_util.ssl_fallback_aiohttp_client_session(
                commons_constants.KNOWN_POTENTIALLY_SSL_FAILED_REQUIRED_URL
            ):
                assert isinstance(session, aiohttp.ClientSession)
                assert len(ok_get_mock_calls) == 0
                assert len(ko_get_mock) == 1
                where_mock.assert_called_once()


async def test_fetch_test_url_with_and_without_certify():
    base_session = aiohttp.ClientSession()
    certify_session = aiohttp_util._get_certify_aiohttp_client_session()
    try:
        async with base_session.get(commons_constants.KNOWN_POTENTIALLY_SSL_FAILED_REQUIRED_URL) as resp:
            assert resp.status < 400
            base_text = await resp.text()
            assert "DrakkarSoftware" in base_text
        async with certify_session.get(commons_constants.KNOWN_POTENTIALLY_SSL_FAILED_REQUIRED_URL) as resp:
            assert resp.status < 400
            certifi_text = await resp.text()
            assert base_text == certifi_text
    finally:
        if base_session:
            await base_session.close()
        if certify_session:
            await certify_session.close()


async def test_certify_aiohttp_client_session():
    origin_where = certifi.where

    with mock.patch.object(certifi, "where", mock.Mock(side_effect=origin_where)) as where_mock:
        async with aiohttp_util.certify_aiohttp_client_session() as session:
            assert isinstance(session, aiohttp.ClientSession)
            where_mock.assert_called_once()


async def test_counter_client_session():
    async with aiohttp_util.CounterClientSession("test") as session:
        assert isinstance(session, aiohttp.ClientSession)
        assert session.per_min.name == "test"
        assert session.per_hour.name == "test"
        assert session.per_day.name == "test"

        # ensure counters work
        async with session.get("https://ifconfig.me/", headers={"Content-Type": "application/json"}) as resp:
            assert resp.status == 200
            assert session.per_min.paths == {"[GET] /": 1}
            assert session.per_hour.paths == {"[GET] /": 1}
            assert session.per_day.paths == {"[GET] /": 1}



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import math
import typing


def round_into_str_with_max_digits(number: float, digits_count: int) -> str:
    """
    Round the number with digits_count
    :param number: the number to round
    :param digits_count: the digit count
    :return: the rounded number
    """
    return "{:.{}f}".format(round(number, digits_count), digits_count)


def round_into_float_with_max_digits(number: float, digits_count: int) -> float:
    """
    Round the float number with digits_count
    :param number: the number to round
    :param digits_count: the digit count
    :return: the rounded number
    """
    return float(
        round_into_str_with_max_digits(number=number, digits_count=digits_count)
    )


def get_digits_count(value: typing.Union[float, decimal.Decimal]):
    """
    :return: the number of digits in the given number
    """
    return round(abs(math.log(value, 10)))



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_commons.symbols
import octobot_commons.enums


@pytest.fixture
def spot_symbol():
    return octobot_commons.symbols.Symbol("BTC/USDT")


@pytest.fixture
def perpetual_future_symbol():
    return octobot_commons.symbols.Symbol("BTC/USDT:BTC")


@pytest.fixture
def future_symbol():
    return octobot_commons.symbols.Symbol("ETH/USDT:USDT-210625")


@pytest.fixture
def option_symbol():
    return octobot_commons.symbols.Symbol("ETH/USDT:USDT-211225-40000-C")


@pytest.fixture
def put_option_symbol():
    return octobot_commons.symbols.Symbol("BTC/USDT:BTC-211225-60000-P")


def test_parse_spot_symbol(spot_symbol):
    assert spot_symbol.base == "BTC"
    assert spot_symbol.quote == "USDT"
    assert spot_symbol.settlement_asset == spot_symbol.identifier == spot_symbol.strike_price == ""
    assert spot_symbol.option_type is None


def test_parse_perpetual_future_symbol(perpetual_future_symbol):
    assert perpetual_future_symbol.base == "BTC"
    assert perpetual_future_symbol.quote == "USDT"
    assert perpetual_future_symbol.settlement_asset == "BTC"
    assert perpetual_future_symbol.strike_price == ""
    assert perpetual_future_symbol.option_type is None


def test_parse_future_symbol(future_symbol):
    assert future_symbol.base == "ETH"
    assert future_symbol.quote == "USDT"
    assert future_symbol.settlement_asset == "USDT"
    assert future_symbol.identifier == "210625"
    assert future_symbol.strike_price == ""
    assert future_symbol.option_type is None


def test_parse_option_symbol(option_symbol):
    assert option_symbol.base == "ETH"
    assert option_symbol.quote == "USDT"
    assert option_symbol.settlement_asset == "USDT"
    assert option_symbol.identifier == "211225"
    assert option_symbol.strike_price == "40000"
    assert option_symbol.option_type == octobot_commons.enums.OptionTypes.CALL.value


def test_base_and_quote(spot_symbol, option_symbol):
    assert spot_symbol.base_and_quote() == ("BTC", "USDT")
    assert option_symbol.base_and_quote() == ("ETH", "USDT")


def test_is_linear(spot_symbol, perpetual_future_symbol,  option_symbol):
    assert spot_symbol.is_linear() is True
    assert perpetual_future_symbol.is_linear() is False
    assert option_symbol.is_linear() is True


def test_is_inverse(spot_symbol, perpetual_future_symbol, option_symbol):
    assert spot_symbol.is_inverse() is False
    assert perpetual_future_symbol.is_inverse() is True
    assert option_symbol.is_inverse() is False


def test_merged_str_symbol_with_full_option(option_symbol, put_option_symbol):
    call_symbol = octobot_commons.symbols.Symbol("ETH/USDT:USDT-211225-40000-C")
    call_symbol.option_type = octobot_commons.enums.OptionTypes.CALL.value
    call_symbol.strike_price = 40000
    call_symbol.identifier = "211225"
    assert call_symbol.merged_str_symbol() == "ETH/USDT:USDT-211225-40000-C"
    
    put_symbol = octobot_commons.symbols.Symbol("BTC/USDT:BTC-211225-60000-P")
    put_symbol.option_type = octobot_commons.enums.OptionTypes.PUT.value
    put_symbol.strike_price = 60000
    put_symbol.identifier = "211225"
    assert put_symbol.merged_str_symbol() == "BTC/USDT:BTC-211225-60000-P"

    custom_symbol = octobot_commons.symbols.Symbol("BTC/USDT:BTC-211225-60000-yES")
    custom_symbol.option_type = "YES"
    custom_symbol.strike_price = 60000
    custom_symbol.identifier = "211225"
    assert custom_symbol.merged_str_symbol() == "BTC/USDT:BTC-211225-60000-YES"


def test_is_put_option():
    put_symbol = octobot_commons.symbols.Symbol("BTC/USDT:BTC-211225-60000-P")
    assert put_symbol.is_put_option() is True
    assert put_symbol.is_call_option() is False


def test_is_call_option():
    call_symbol = octobot_commons.symbols.Symbol("ETH/USDT:USDT-211225-40000-C")
    assert call_symbol.is_call_option() is True
    assert call_symbol.is_put_option() is False


def test_is_put_and_call_option_with_non_option_symbols(spot_symbol, perpetual_future_symbol, future_symbol):
    assert spot_symbol.is_put_option() is False
    assert spot_symbol.is_call_option() is False
    assert perpetual_future_symbol.is_put_option() is False
    assert perpetual_future_symbol.is_call_option() is False
    assert future_symbol.is_put_option() is False
    assert future_symbol.is_call_option() is False


def test_does_expire(spot_symbol, perpetual_future_symbol, future_symbol, option_symbol, put_option_symbol):
    assert spot_symbol.does_expire() is False
    assert perpetual_future_symbol.does_expire() is False
    assert future_symbol.does_expire() is True
    assert option_symbol.does_expire() is True
    assert put_option_symbol.does_expire() is True


def test__eq__(spot_symbol, option_symbol):
    assert spot_symbol == octobot_commons.symbols.Symbol("BTC/USDT")
    assert spot_symbol != octobot_commons.symbols.Symbol("BTC/USD")
    assert spot_symbol != option_symbol
    assert option_symbol == octobot_commons.symbols.Symbol("ETH/USDT:USDT-211225-40000-C")
    assert option_symbol != octobot_commons.symbols.Symbol("ETH/USDT:USDT-211225-40000-P")



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import pytest

import octobot_commons.tree as tree
import octobot_commons.asyncio_tools as asyncio_tools


@pytest.fixture
def event_tree():
    return tree.EventTree()


@pytest.fixture
def event_tree_node():
    return tree.EventTreeNode()


def test_event_tree_create_node_at_path(event_tree):
    assert event_tree.root.children == {}
    event_tree.create_node_at_path(["1"], True)
    assert isinstance(event_tree.get_node(["1"]), tree.EventTreeNode)
    assert event_tree.get_node(["1"]).is_triggered()
    event_tree.create_node_at_path(["2", "abtc"], False)
    assert not event_tree.get_node(["2", "abtc"]).is_triggered()
    assert isinstance(event_tree.get_node(["1"]), tree.EventTreeNode)
    assert isinstance(event_tree.get_node(["2", "abtc"]), tree.EventTreeNode)


def test_event_tree_node_trigger(event_tree_node):
    assert not event_tree_node.is_triggered()
    event_tree_node.trigger()
    assert event_tree_node.is_triggered()


def test_event_tree_node_bind_parent(event_tree_node):
    parent_node = tree.EventTreeNode()
    event_tree_node.bind_parent(parent_node)
    assert event_tree_node.get_parent() is parent_node
    parent_node.children["d"] = event_tree_node
    assert not event_tree_node.is_triggered()
    assert not parent_node.is_triggered()
    event_tree_node.trigger()
    assert event_tree_node.is_triggered()
    assert parent_node.is_triggered()


@pytest.mark.asyncio
async def test_event_tree_node_wait(event_tree_node):
    waiter = asyncio.create_task(asyncio.wait_for(event_tree_node.wait(), 0.1))
    for _ in range(10):
        # let async loop end task if necessary
        await asyncio_tools.wait_asyncio_next_cycle()
    assert not event_tree_node.is_triggered()
    assert not waiter.done()
    event_tree_node.trigger()
    assert not waiter.done()
    # need 2 cycles to both end the waiter and end the wait_for
    for _ in range(2):
        await asyncio_tools.wait_asyncio_next_cycle()
    assert waiter.done()


def test_event_tree_node_set_child(event_tree_node):
    assert event_tree_node.children == {}
    node = tree.EventTreeNode()
    event_tree_node.set_child("hi", node)
    assert event_tree_node.children == {
        "hi": node
    }


def test_event_tree_node_pop_child(event_tree_node):
    assert event_tree_node.children == {}
    node = tree.EventTreeNode()
    event_tree_node.set_child("hi", node)
    assert event_tree_node.pop_child("hi", None) is node
    assert event_tree_node.children == {}
    assert event_tree_node.pop_child("hi", None) is None
    assert event_tree_node.children == {}


def test_event_tree_node_on_child_change(event_tree_node):
    assert not event_tree_node.is_triggered()
    event_tree_node.on_child_change()
    # no child, need manual trigger
    assert not event_tree_node.is_triggered()
    event_tree_node.trigger()
    assert event_tree_node.is_triggered()
    event_tree_node.on_child_change()
    # still triggered
    assert event_tree_node.is_triggered()

    parent_node = tree.EventTreeNode()
    parent_node.children["a"] = event_tree_node
    assert not parent_node.is_triggered()
    event_tree_node.bind_parent(parent_node)
    assert parent_node.is_triggered()

    event_tree_node.clear()
    assert not event_tree_node.is_triggered()
    assert not parent_node.is_triggered()

    other_child = tree.EventTreeNode()
    parent_node.children["b"] = other_child
    other_child.bind_parent(parent_node)
    assert not parent_node.is_triggered()

    event_tree_node.trigger()
    assert not other_child.is_triggered()
    assert not parent_node.is_triggered()

    other_child.trigger()
    assert event_tree_node.is_triggered()
    assert parent_node.is_triggered()

    other_child.clear()
    assert event_tree_node.is_triggered()
    assert not parent_node.is_triggered()


def test_event_tree_node_get_path_to_root(event_tree_node):
    assert event_tree_node.children == {}
    assert event_tree_node.get_path_to_root() == []
    parent = tree.EventTreeNode()
    event_tree_node.bind_parent(parent)
    parent.set_child("hi", event_tree_node)
    assert event_tree_node.get_path_to_root() == ["hi"]
    parent_2 = tree.EventTreeNode()
    parent.bind_parent(parent_2)
    parent_2.set_child("hello", parent)
    assert event_tree_node.get_path_to_root() == ["hello", "hi"]
    assert parent.get_path_to_root() == ["hello"]
    assert parent_2.get_path_to_root() == []
    other_child = tree.EventTreeNode()
    other_child.bind_parent(parent)
    parent.set_child("ho", other_child)
    assert event_tree_node.get_path_to_root() == ["hello", "hi"]
    assert other_child.get_path_to_root() == ["hello", "ho"]


def test_event_tree_node_get_child_key(event_tree_node):
    with pytest.raises(KeyError):
        event_tree_node.get_child_key("dd")
    other = tree.EventTreeNode()
    event_tree_node.children["dd"] = other
    assert event_tree_node.get_child_key(other) == "dd"



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.dsl_interpreter.operator as dsl_interpreter_operator


class UnaryOperator(
    dsl_interpreter_operator.Operator
):  # pylint: disable=abstract-method
    """
    Base class for unary operators.
    Unary operators have one operand.
    """

    def __init__(
        self,
        operand: dsl_interpreter_operator.OperatorParameterType,
        **kwargs: typing.Any
    ):
        """
        Initialize the unary operator with its operand.
        """
        super().__init__(operand, **kwargs)

    def get_computed_operand(
        self,
    ) -> dsl_interpreter_operator.ComputedOperatorParameterType:
        """
        Get the computed operand of the unary operator.
        """
        computed_parameters = self.get_computed_parameters()
        return computed_parameters[0]



# Copyright
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_commons.signals import signal

from octobot_commons.signals.signal import (
    Signal,
)


from octobot_commons.signals import signal_dependencies

from octobot_commons.signals.signal_dependencies import (
    SignalDependencies,
)

from octobot_commons.signals import signal_bundle

from octobot_commons.signals.signal_bundle import (
    SignalBundle,
)

from octobot_commons.signals import signal_bundle_builder

from octobot_commons.signals.signal_bundle_builder import (
    SignalBundleBuilder,
)

from octobot_commons.signals import signal_factory

from octobot_commons.signals.signal_factory import (
    create_signal_bundle,
    create_signal,
)

from octobot_commons.signals import signals_emitter

from octobot_commons.signals.signals_emitter import (
    emit_signal_bundle,
)

from octobot_commons.signals import signal_builder_wrapper

from octobot_commons.signals.signal_builder_wrapper import (
    SignalBuilderWrapper,
)

from octobot_commons.signals import signal_publisher

from octobot_commons.signals.signal_publisher import (
    SignalPublisher,
)


__all__ = [
    "Signal",
    "SignalBundle",
    "create_signal_bundle",
    "create_signal",
    "emit_signal_bundle",
    "SignalBuilderWrapper",
    "SignalPublisher",
    "SignalDependencies",
]
